.( ftra.fs ) cr
\ -----------------------------

\ Copyright (C) 2005,2007,2013 Marcos Cruz (programandala.net)
\ License: http://programandala.net/license

\ Text strings translation tool
\ Program written in Forth; developed with Forth 5mx and Gforth.

\ History
\ ------------------------------------------

\ 2005-11-01
\   Start. Algorithm to translate chars into strings.
\ 2005-11-10
\   Rewrite. New improved algorithm to translate strings into strings.
\ 2005-11-11
\   Bugs fixed. It seems it works.
\ 2007-05-08
  \ 'translate-more?' factored out from 'tra'.
\   Bug fixed in 'tra' :
\   The translation halted when the last word in the text had to be translated but
\   it was shorter than the first word in the translation table.
\   The best order for the elements in the translation table is not solved yet.
\   Provisional solution: remove that condition.
\ 2013-12-10
\   Source code style and stack notation updated.
\   '-left' removed, '/string' does exactly the same.
\   '.tra-table-element' factored from '.tra-table'.
\   Some comments fixed.

\ ------------------------------------------

marker ftra

variable tra-table  \ address of the current translation table
variable tra-pointer  \ pointer to the current string in the current translation table

: tra-pointer0  ( -- )
  \ Init the translation pointer.
  tra-table @ tra-pointer !
  ;
: tra-table:  ( "<spaces>name" -- )
  \ Start the translation table definition.
  create  does>  ( -- ) ( dfa ) tra-table ! tra-pointer0
  ;
: ;tra-table  ( -- )
  \ Finish the translation table definition.
  0 dup c, c, align
  ;
: c>s,  ( b -- )
  \ Compile a one char string into the data space.
  1 c, c,
  ;
variable translation  \ address of the translation
variable >translation  \ pointer to the next char in the translation
: translation+!  ( ca len -- )
  \ Add a string to the current translation.
  dup >r >translation @ swap cmove  \ move chars
  r> dup >translation +!  \ update char pointer
  translation @ dup C@ rot + swap C!  \ update string count
  ;
: translate-this?  ( ca1 len1 ca2 len2 -- ca1 len1 ca2 len2 wf )
  \ Is ca2 len2 a prefix of ca1 len1?
  \ ca1 len1 = remaining text to be translated
  \ ca2 len2 = text to translate, from the translation table
  dup >r 2over 2over 2swap drop r> compare 0=
  ;
: tra-pointer@  ( -- ca len )
  \ Fetch the string pointed.
  tra-pointer @ count
  ;
: tra-pointer+  ( -- )
  \ Update tra-pointer to the next string in the translation table.
  tra-pointer dup @ dup C@ 1+ swap + swap !
  ;
: translation0  ( -- )
  \ Init the translation.
  here 
  dup translation !  \ address
  0 over c!  \ char count
  1+ >translation !  \ pointer to the first char
  ;
: translate-more?  ( ca1 len1 -- ca1 len1 ca2 len2 wf )
  \ Keep on translating?
  \ ca1 len1 = remaining text to be translated
  \ ca2 len2 = text to translate, from the translation table
  \ wf = translate the text?
  [ false ] [if]
    \ Original version with 3 conditions, and debug code.
    \ dup 
    dup 0<> \ is there text to translate?
\    cr dup if ." C1" then \ xxx informer
    >r tra-pointer@ 
    ( ca1 len1 ca2 len2 -- )
    \ ca1 len1 = remaining text to be translated
    \ ca2 len2 = text to translate, from the translation table
    dup 0<>  \ is there a possible translation?
\    dup if ." C2" then \ xxx informer
    r> and  
    \ over r> <=  \ is the possible translation shorter than the text to translate?
\    dup if ." C3" then  \ xxx informer
    \ and 
  [else]
    \ 2007-05-08 New version with only 2 conditions.
    dup 0<> >r \ is there text to translate?
    tra-pointer@ dup 0<>  \ is there a possible translation?
    r> and 
  [then]
  ;
: tra  ( ca1 len1 -- ca2 len2 )
  \ Translate a string using the current translation table.
  \ ca1 len1 = text to be translated
  \ ca2 len2 = translated text
\ 2dup cr type ." --> ? " \ xxx informer
  translation0
  begin
    dup
\    more and \ xxx informer
  while  tra-pointer0
    begin  translate-more?
    while
\      cr ."  translate more: " 2over type ."  --> " 2dup type  \ xxx informer
      tra-pointer+ translate-this?
      if      
\      cr ."  translate this:"  2over type ."  --> " 2dup type \ xxx informer
        nip /string  tra-pointer@ translation+!  tra-pointer0
      else
\      cr ."  don't translate this: " 2over type ."  --> " 2dup type  \ xxx informer
        2drop  tra-pointer+
      then
    repeat  2drop dup
    if
      over 1 translation+!  1 /string  \ char not translated
    then
  repeat  2drop  translation @ count
\  2dup cr ." RESULT: " type \ xxx informer
  ;
: .tra-table-element  ( ca len -- )
  \ Print the content of an element of the current translation table
  cr type ."  > "  tra-pointer+
  tra-pointer@ type  tra-pointer+
  ;
: .tra-table  ( -- )
  \ Print the content of the current translation table
  tra-pointer0
  begin   tra-pointer@ dup
  while   .tra-table-element
  repeat  2drop
  ;

.(  ftra.fs compiled ) cr
